/*
   event horizon - lua/c++ event system demonstration
   Copyright (c) 2006 kevin meinert all rights reserved

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301  USA
   */

/** @file EventSystem.h
 *   defines an event system useful for games
 *   with an event notification mechanism compatible with both lua and c/c++ method callbacks
 *   each event notification can deliver a packet of information (a data payload) 
 *        that may be used as arguments (or context) for the event.
 *   this payload mechanism brings much variability to each individual event type
 *   event types are defined as fast memory-aligned short-strings (128 bit), and is up to the user to uniqify
 */

#ifndef EVENTSYSTEM_H
#define EVENTSYSTEM_H

#include <map>
#include "packet.h"
#include "ShortString.h"
#include "FastDelegate.h" // generic function pointer holder...

#ifndef ASSERT
#define ASSERT( a,b ) assert( a && b );
#endif

/// simple structure for recording (one) lua function argument when lua calls into C/C++
struct LuaScriptArg
{
   // preallocate string, to avoid hitting mem allocs very often
   // (assuming LuaScriptData isnt add/removing LuaScriptArg's either)
   // if strings need more memory, size will expand, and next time still be allocated to avoid future allocs..
   LuaScriptArg() { s.reserve( 1024 ); }
   enum LuaScriptArgType
   {
      TYPEINT, TYPESTR, TYPEFLOAT, TYPEBOOL
   };
   LuaScriptArgType type;
   union
   {
      int i;
      float f;
      bool b;
   };
   std::string s;
};

/// simple structure for recording (a list of) lua function arguments when lua calls into C/C++
/// preallocates buffer of args to eliminate hitting new/delete.
class LuaScriptData
{
public:
   LuaScriptData( unsigned int size )
   {
      mSize = 0;
      mData.resize( size );
   }

   /// array interface...
   inline unsigned int size() const { return mSize; }
   inline void push_back( const LuaScriptArg& arg ) { ASSERT( mSize < mData.size(), "too many args, increase" ); mData[mSize] = arg; ++mSize; }
   inline void clear() { mSize = 0; }
   inline LuaScriptArg& operator[]( unsigned int x ) { ASSERT( x < mSize, "out of bounds" ); return mData[x]; }
   inline const LuaScriptArg& operator[]( unsigned int x ) const { ASSERT( x < mSize, "out of bounds" ); return mData[x]; }

   /// push a vector of scriptdata onto the lua stack
   inline void pushLuaData( lua_State* L )
   {
      const LuaScriptData& scriptdata = *this;
      for (unsigned int x = 1; x < scriptdata.size(); ++x)
      {
         switch (scriptdata[x].type)
         {
         case LuaScriptArg::TYPEINT:
            lua_pushinteger( L, scriptdata[x].i );
            break;
         case LuaScriptArg::TYPEBOOL:
            lua_pushboolean( L, scriptdata[x].b );
            break;
         case LuaScriptArg::TYPEFLOAT:
            lua_pushnumber( L, scriptdata[x].f );
            break;
         case LuaScriptArg::TYPESTR:
            lua_pushstring( L, scriptdata[x].s.c_str() );
            break;
         }
      }
   }

   /// convert scriptdata vector into a packed C/C++ structure as defined by the mArgsDescription packet.
   inline void convertToCargs( char* cargs, unsigned int cargsBytes, 
                                          const packet& argsDescription )
   {
      ASSERT( argsDescription.size <= cargsBytes, "not enough space in cargs, expand then recompile" );
      const LuaScriptData& scriptdata = *this;
      for (unsigned int x = 0; x < argsDescription.attribs.size(); ++x)
      {
         const size_t end = (argsDescription.attribs[x].location + argsDescription.attribs[x].size);
         ASSERT( end <= argsDescription.size, "cargs not big enough" );
         switch (argsDescription.attribs[x].type)
         {
         case packet::DataMember::TYPEBOOL:
            ASSERT( scriptdata[x+1].type == LuaScriptArg::TYPEFLOAT, "wrong type from lua, bool expected" );
            (*(bool*)&cargs[argsDescription.attribs[x].location]) = (scriptdata[x+1].f < 0.001f ? false : true);
            break;
         case packet::DataMember::TYPEINT:
            ASSERT( scriptdata[x+1].type == LuaScriptArg::TYPEFLOAT, "wrong type from lua, int expected" );
            (*(int*)&cargs[argsDescription.attribs[x].location]) = (int)scriptdata[x+1].f;
            break;
         case packet::DataMember::TYPESTR:
            ASSERT( scriptdata[x+1].type == LuaScriptArg::TYPESTR, "wrong type from lua, str expected" );
            strncpy( &cargs[argsDescription.attribs[x].location], scriptdata[x+1].s.c_str(), argsDescription.attribs[x].size );
            break;
         case packet::DataMember::TYPEFLOAT:
            ASSERT( scriptdata[x+1].type == LuaScriptArg::TYPEFLOAT, "wrong type from lua, float expected" );
            (*(float*)&cargs[argsDescription.attribs[x].location]) = scriptdata[x+1].f;
            break;
         }
      }
   }
private:
   std::vector<LuaScriptArg> mData;
   unsigned int mSize;
};


/// represents a lua or C++ function callback
/// to subscribe to an event, register an EventCallback with the EventSystem 
struct EventCallback
{
   typedef void CreturnType;
   typedef const void* CparamType;
   typedef fastdelegate::FastDelegate1<CparamType> CfunctionType;

   /// default constructor
   EventCallback() : mLuaFuncReference( -1 ), mLuaObjReference( -1 ) {}
   
   /// function pointer constructor
   EventCallback( CreturnType (*function_to_bind)( CparamType p1 ), const packet& argsdesc ) : mLuaFuncReference( -1 ), mLuaObjReference( -1 )
   {
      initFromCfunction( CfunctionType( function_to_bind ), argsdesc );
   }

   /// non-static member function pointer constructor (for single, virtual, or multiple)
   template <class OBJTYPEFP, class OBJTYPE>
	EventCallback( OBJTYPE* pthis, CreturnType (OBJTYPEFP::* function_to_bind)( CparamType p1 ), const packet& argsdesc ) : mLuaFuncReference( -1 ), mLuaObjReference( -1 )
   {
      initFromCfunction( CfunctionType( pthis, function_to_bind ), argsdesc );
   }

   /// const non-static member function pointer constructor (for single, virtual, or multiple)
   template <class OBJTYPEFP, class OBJTYPE>
	EventCallback( const OBJTYPE* pthis, CreturnType (OBJTYPEFP::* function_to_bind)( CparamType p1 ) const, const packet& argsdesc ) : mLuaFuncReference( -1 ), mLuaObjReference( -1 )
   {
      initFromCfunction( CfunctionType( pthis, function_to_bind ), argsdesc );
   }

   /// set the Lua callback function
   /// supports 2 types of function callback:
   /// function: 
   ///   expects a <function> typed object on the lua stack before calling this method
   /// obj,member pair
   ///   expects <function> <table> on the lua stack
   /// automatically determines which type of callback by looking at what's on the stack...
   void initFromLuaFunction( lua_State* L )
   {
      // easy way (was getting collisions with LUA_REGISTRYINDEX, worked ok with GLOBALS).
      //mLuaFuncReference = luaL_ref( L, LUA_REGISTRYINDEX );

      // store function (from the stack) into callbacks[mLuaFuncReference]
      static bool maketable = true;
      if (maketable)
      {
         maketable = false;
         lua_newtable( L );
         lua_setfield(L, LUA_REGISTRYINDEX, "callbacks" );
      }

      // support registering normal functions, or obj,memberfunc pairs...
      if (0 < lua_gettop(L) && lua_type(L, -1) == LUA_TTABLE)
      {
         lua_getfield(L, LUA_REGISTRYINDEX, "callbacks" ); // push table "callbacks"
         lua_insert(L, -2);                     // swap  table and function
         mLuaObjReference = luaL_ref( L, -2 );   // pop object into callbacks[mLuaObjReference]
         lua_pop( L, 1 );                       // pop off the "callbacks" table
      }
      if (0 < lua_gettop(L) && lua_type(L, -1) == LUA_TFUNCTION)
      {
         lua_getfield(L, LUA_REGISTRYINDEX, "callbacks" ); // push table "callbacks"
         lua_insert(L, -2);                     // swap  table and function
         mLuaFuncReference = luaL_ref( L, -2 );  // pop function into callbacks[mLuaFuncReference]
         lua_pop( L, 1 );                       // pop off the "callbacks" table
      }
   }

   /// set C callback function
   void initFromCfunction( CfunctionType func, const packet& argsdesc )
   {
      mCFuncPtr = func;
      mArgsDescription = argsdesc;
   }

   /// after initFromLuaFunction has been used, call pushLuaFunc to push the function back onto the stack.
   /// usually call pushLuaFunc, then push lua args, then call lua_pcall to call the function.
   /// @result returns number of additional args pushed onto the stack (i.e. for self when callback is an obj,member)
   unsigned int pushLuaFunc( lua_State* L )
   {
      // easy way (didn't work well, see above)
      //lua_rawgeti( L, LUA_REGISTRYINDEX, mLuaFuncReference );

      // retrieve the lua function from callbacks[mLuaFuncReference]
      lua_getfield(L, LUA_REGISTRYINDEX, "callbacks" ); // push table "callbacks"
      lua_rawgeti( L, -1, mLuaFuncReference ); // push function from callbacks[mLuaFuncReference]
      lua_insert(L, -2);                      // swap table and function
      lua_pop( L, 1 );                        // pop table
      if (!lua_isfunction( L, -1 ))
      {
         std::cout << "error: couldn't retrieve lua function callback" << std::endl;
      }

      // also push 'self' if this is an obj,member pair
      if (0 <= mLuaObjReference)
      {
         // retrieve the lua table from callbacks[mLuaObjReference]
         lua_getfield(L, LUA_REGISTRYINDEX, "callbacks" ); // push table "callbacks"
         lua_rawgeti( L, -1, mLuaObjReference );  // push table from callbacks[mLuaObjReference]
         lua_insert(L, -2);                      // swap table and function
         lua_pop( L, 1 );                        // pop table
         if (!lua_istable( L, -1 ))
         {
            std::cout << "error: couldn't retrieve self object for lua function callback" << std::endl;
         }
         return 1;
      }
      return 0;
   }

   const packet& getCArgsDesc() const { return mArgsDescription; }
   CfunctionType getCFunc() { return mCFuncPtr; }

   /// is this callback to lua, or to c/c++?
   inline bool isLuaFunc() const { return mLuaFuncReference != -1; }
protected:
   // lua callback
   friend class EventSubscriberRegistry;
   int mLuaFuncReference, mLuaObjReference;

   // c function delegate
   CfunctionType mCFuncPtr;
   // struct description for the c function argument (arg can vary from callback to callback)
   packet mArgsDescription;
};

/// database of functions to be called when events are triggered.
class EventSubscriberRegistry
{
public:
   typedef std::vector<EventCallback> EventSubscriberCallbacks;
   typedef std::map<ShortString16, EventSubscriberCallbacks> EventSubscribers;

   /// add a callback for eventname
   void add( const ShortString16& eventname, const EventCallback& callback )
   {
      mEventSubscribers[eventname].push_back( callback );
   }

   /// get callbacks for event name.
   EventSubscriberCallbacks& getCallbacks( const ShortString16& eventname )
   {
      return mEventSubscribers[eventname];
   }

   /// debug print
   void debugprint()
   {
      EventSubscribers::iterator mapit;
      for (mapit = mEventSubscribers.begin(); 
            mapit != mEventSubscribers.end();
            ++mapit)
      {
         EventSubscriberCallbacks& vec = (*mapit).second;
         for (unsigned int x = 0; x < vec.size(); ++x)
         {
            char buf[24];
            (*mapit).first.get( buf, 24 );
            std::cout << "[" << x << "] name:" << buf;
            if (0 <= (*mapit).second[x].mLuaObjReference)
               std::cout << " luaobj_id:" << (*mapit).second[x].mLuaObjReference;
            if (0 <= (*mapit).second[x].mLuaFuncReference)
               std::cout << " luafunc_id:" << (*mapit).second[x].mLuaFuncReference;
            std::cout << " " << std::endl;
         }
      }
   }

private:
   EventSubscribers mEventSubscribers; //< functions registered to be called when events are triggered.
};


/// event system.
class EventSystem
{
public:
   /// call this to initialize the EventSystem object.
   void init( lua_State* state )
   {
      L = state;
      bind( L );
   }

   /// call this to register a function to be called when an event happens
   /// @code
   ///   void func( void* args ) { MyArg* a = (MyArg*)args; /* do stuff */ }
   ///   reg( "fire", EventCallback( &func, MyArg::getdef() ) );
   /// @endcode
   void reg( const ShortString16& eventname, EventCallback& callback )
   {
      mEventSubscriberRegistry.add( eventname, callback );

      char str[24];
      eventname.get( str, 24 );
      std::cout << "registered [c++] callback for " << str << " event" << std::endl;
      mEventSubscriberRegistry.debugprint();
   }

   /// call this to trigger an event
   /// @code
   ///    trigger( "fire", MyArgs( 1, 64, "four" ) );
   /// @endcode
   template <typename T>
   inline void trigger( const ShortString16& eventname, const T& args )
   {
      // you must have getdef defined for your args structure, see packet_ for details.
      trigger( eventname, (char*)&args, T::getdef() );
   }

   /// get the database of event subscriber callbacks...
   EventSubscriberRegistry& getEventSubscriberRegistry()
   {
      return mEventSubscriberRegistry;
   }

   /// this class can be instantiated for special uses, but there is also one 
   /// global instance... most systems will use this one instance.
   static EventSystem& instance() { static EventSystem es; return es; }

private:
   /// exposes 2 lua script functions: trigger and reg
   /// reg:      reg( "eventname", func )
   /// trigger:  trigger( "eventname", arg1, arg2, arg3, ... )
   static void bind( lua_State* L )
   {
      lua_pushcfunction( L, EventSystem::trigger );
      lua_setglobal( L, "trigger" );
      lua_pushcfunction( L, EventSystem::reg );
      lua_setglobal( L, "reg" );
   }

   /// private func called by trigger c-interface ...
   void trigger( const ShortString16& eventname, char* cargs, const packet& desc )
   {
      static LuaScriptData scriptdata( 20 ); //< # of slots for incoming function arguments
      scriptdata.clear();

      scriptdata.push_back( LuaScriptArg() ); // 1st param ignored...

      for (unsigned int x = 0; x < desc.attribs.size(); ++x)
      {
         switch (desc.attribs[x].type)
         {
         case packet::DataMember::TYPEBOOL:
            {
               LuaScriptArg dummy;
               dummy.f = (float)(*(bool*)&cargs[desc.attribs[x].location]);
               dummy.type = LuaScriptArg::TYPEFLOAT;
               scriptdata.push_back( dummy );
               break;
            }
         case packet::DataMember::TYPEINT:
            {
               LuaScriptArg dummy;
               dummy.f = (float)(*(int*)&cargs[desc.attribs[x].location]);
               dummy.type = LuaScriptArg::TYPEFLOAT;
               scriptdata.push_back( dummy );
               break;
            }
         case packet::DataMember::TYPESTR:
            {
               LuaScriptArg dummy;
               dummy.s = (char*)&cargs[desc.attribs[x].location];
               dummy.type = LuaScriptArg::TYPESTR;
               scriptdata.push_back( dummy );
               break;
            }
         case packet::DataMember::TYPEFLOAT:
            {
               LuaScriptArg dummy;
               dummy.f = (*(float*)&cargs[desc.attribs[x].location]);
               dummy.type = LuaScriptArg::TYPEFLOAT;
               scriptdata.push_back( dummy );
               break;
            }
         }
      }

      EventSubscriberRegistry::EventSubscriberCallbacks& callbacks = getEventSubscriberRegistry().getCallbacks( eventname );
      for (unsigned int x = 0; x < callbacks.size(); ++x)
      {
         // call any registered lua callbacks
         if (callbacks[x].isLuaFunc())
         {
            unsigned int args = callbacks[x].pushLuaFunc( L );
            scriptdata.pushLuaData( L );
            const int nargs = (int)scriptdata.size() - 1 + args;
            lua_pcall( L, nargs, 0, 0 );
         }
         else
         {
            callbacks[x].getCFunc()( cargs ); // call
         }
      }
   }

   /// Lua version of reg() - subscribe to an event, using a lua function as callback
   /// arg1 = eventname (shortstring, 16 chars or less)
   /// arg2 = function (lua function)
   /// @see bind()
   static int reg( lua_State* L )
   {
      bool functionReg = (2 == lua_gettop(L) && lua_type(L, -2) == LUA_TSTRING && lua_type(L, -1) == LUA_TFUNCTION);
      bool classMemberReg = (3 == lua_gettop(L) && lua_type(L, -3) == LUA_TSTRING && lua_type(L, -2) == LUA_TFUNCTION && lua_type(L, -1) == LUA_TTABLE);
      if (functionReg || classMemberReg)
      {
         const char* str = lua_tostring( L, functionReg ? -2 : -3 ); // read the eventname
         assert( strlen( str ) <= 16 && "shortstring, 16 chars or less" );
         ShortString16 eventname = str;
         EventCallback callback;
         callback.initFromLuaFunction( L );
         lua_pop( L, 1 );                        // pop off the eventname
         instance().getEventSubscriberRegistry().add( eventname, callback );
         std::cout << "registered [lua] callback for " << str << " event" << std::endl;
         instance().getEventSubscriberRegistry().debugprint();
         ASSERT( 0 == lua_gettop(L), "something's left on the stack" );
      }
      else
      {
         std::cout << "error: not a function: usage reg( \"eventname\", func, <optional obj> )" << std::endl;
         //BR_DUMP_STACK( L, "blah" );
      }
      return 0;
   }

   /// lua version of trigger() - trigger an event
   /// usage:  trigger( "eventname", arg1, arg2, arg3 )
   /// @see bind()
   static int trigger( lua_State* L )
   {
      static LuaScriptData scriptdata( 20 ); //< # of slots for incoming function arguments
      scriptdata.clear();

      int top = lua_gettop(L);
      for (int i = 1; i <= top; ++i)
      {
         int t = lua_type(L, i);
         switch (t)
         {
         case LUA_TSTRING:
            {
               LuaScriptArg dummy;
               dummy.s = lua_tostring(L, i);
               dummy.type = LuaScriptArg::TYPESTR;
               scriptdata.push_back( dummy );
            }
            break;
         case LUA_TBOOLEAN:
            {
               LuaScriptArg dummy;
               dummy.b = (0 == lua_toboolean(L, i) ? false : true);
               dummy.type = LuaScriptArg::TYPEBOOL;
               scriptdata.push_back( dummy );
               break;
            }
         case LUA_TNUMBER:
            {
               LuaScriptArg dummy;
               dummy.f = (float)lua_tonumber(L, i);
               dummy.type = LuaScriptArg::TYPEFLOAT;
               scriptdata.push_back( dummy );
               break;
            }
         }
      }
      lua_pop( L, lua_gettop( L ) );

      // do validation checks on the incoming function arguments
      if (0 < scriptdata.size() && scriptdata[0].type == LuaScriptArg::TYPESTR)
      {
         //std::cout << "trigger: " << scriptdata[0].s << std::endl;
         for (unsigned int x = 1; x < scriptdata.size(); ++x)
         {
            switch (scriptdata[x].type)
            {
            case LuaScriptArg::TYPESTR: /*std::cout << scriptdata[x].s << std::endl;*/ break;
            case LuaScriptArg::TYPEBOOL: /*std::cout << scriptdata[x].b << std::endl;*/ break;
            case LuaScriptArg::TYPEFLOAT: /*std::cout << scriptdata[x].f << std::endl;*/ break;
            default: std::cout << "unknown param" << std::endl;
            }
         }
      }
      else
      {
         std::cout << "error: trigger called without an event name as 1st argument" << std::endl;
         return 0;
      }

      // dispatch calls to each event subscriber
      ShortString16 eventname = scriptdata[0].s.c_str();
      EventSubscriberRegistry::EventSubscriberCallbacks& callbacks = instance().getEventSubscriberRegistry().getCallbacks( eventname );
      for (unsigned int x = 0; x < callbacks.size(); ++x)
      {
         // call any registered lua callbacks
         if (callbacks[x].isLuaFunc())
         {
            unsigned int args = callbacks[x].pushLuaFunc( L );
            scriptdata.pushLuaData( L );
            const int nargs = (int)scriptdata.size() - 1 + args;
            lua_pcall( L, nargs, 0, 0 );
         }
         else
         {
            static const size_t MAX_BYTES = 512;
            static char cargs[MAX_BYTES];
            scriptdata.convertToCargs( cargs, MAX_BYTES, callbacks[x].getCArgsDesc() );
            callbacks[x].getCFunc()( cargs ); // call
         }
      }
      return 0;
   }

private:
   EventSubscriberRegistry mEventSubscriberRegistry;
   lua_State* L;
}; // end EventSystem


#endif // EVENTSYSTEM