/*
   event horizon - lua/c++ event system demonstration
   Copyright (c) 2006 kevin meinert all rights reserved

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301  USA
   */

#ifndef SHORT_STRING
#define SHORT_STRING

#include <string.h>
#include <assert.h>

/// ShortString.
/// Description:  A string who's size is fixed and (ideally) memory aligned.  Copys and compares are fast.
/// NOTE: null-terminated unless full size is used, then no null termination.  
///       use get and set to convert to char* strings.
/// 
/// The following types are also defined: 
///		 ShortString8, ShortString16, ShortString32, ShortString64, ShortString128
template <unsigned size = 16>
class ShortString
{
public:
      enum { INTS = size/4, SIZE = size  };
      union
      {
         char szString[SIZE];
         int iData[INTS];
      };

      /// constructor
      inline ShortString()
      {
         for (int x = 0; x < INTS; ++x)
            iData[x] = 0;
      }

      /// copy constructor from another shortstring
      inline ShortString( const ShortString& oString )
      {
         set( oString );
      }

      /// copy constructor from null terminated string
      inline ShortString( const char* const szBuf )
      {
         set( szBuf );
      }

      /// set the ShortString using another ShortString.
      inline void set( const ShortString& oString )
      {
         for (int x = 0; x < INTS; ++x)
         {
            iData[x] = oString.iData[x];
         }
      }

		/// set the ShortString using a null-terminated string.
      /// converts string to ShortString
      inline void set( const char* const szBuf )
      {
         const size_t len = strlen( szBuf );
         const size_t n = (SIZE < len) ? SIZE : len;
         strncpy( szString, szBuf, n );
         if (0 < (SIZE - n))
         {
			   memset( &szString[n], 0, SIZE - n );
         }
      }

      
      
		/// get a null-terminated string from the ShortString.
      /// converts ShortString to string
      inline void get( char* szBuf, int iBufSize ) const
      {
         int len = length();
         int n = (iBufSize < len) ? iBufSize : len;
         strncpy( szBuf, szString, n );
         int terminate = ((iBufSize-1) < len) ? (iBufSize-1) : len;
         szBuf[terminate] = '\0';
      }

		/// get the length of the ShortString, does not include the null termination character (if it exists).
      inline int length() const
      {
         for (int x = 0; x < SIZE; ++x)
         {
            if (szString[x] == '\0') 
               return x;
         }
         return SIZE;
      }

		/// const array operator.  asserts on out of bounds.
      inline const char& operator[]( unsigned int x ) const { assert( 0 <= x && x < SIZE ); return szString[x]; }
      
		/// array operator.  asserts on out of bounds.
      inline char& operator[]( unsigned int x ) { assert( 0 <= x && x < SIZE ); return szString[x]; }

		/// Short String != operator
		template <unsigned SIZE>
		inline bool operator!=( const ShortString<SIZE>& oRhs ) const
		{
			return ! ::operator==( *this, oRhs );
		}

		/// Short String == operator
		template <unsigned SIZE>
		inline bool operator==( const ShortString<SIZE>& oRhs ) const;

		/// lexically compare strings.
		inline bool operator<( const ShortString<SIZE>& oRhs ) const
		{
			for (int x = 0; x < ShortString<SIZE>::INTS; ++x)
			{
				if ( iData[x] < oRhs.iData[x])
               return true;
			}
         return false;
		}

		/// lexically compare strings.
		inline bool operator>( const ShortString<SIZE>& oRhs ) const
		{
			return !(*this < oRhs) && *this != oRhs;
		}

		/// lexically compare strings.
		inline bool operator>=( const ShortString<SIZE>& oRhs ) const
		{
			return *this > oRhs || *this == oRhs;
		}

		/// lexically compare strings.
		inline bool operator<=( const ShortString<SIZE>& oRhs ) const
		{
			return *this < oRhs || *this == oRhs;
		}

      /// set
      inline ShortString& operator=( const ShortString& oString )
      {
         set( oString );
         return *this;
      }
};

typedef ShortString<8> ShortString8;
typedef ShortString<16> ShortString16;
typedef ShortString<32> ShortString32;
typedef ShortString<64> ShortString64;
typedef ShortString<128> ShortString128;

/// check for equality between ShortStrings.
inline static bool operator==( const ShortString<4>& oLhs, const ShortString<4>& oRhs )
{
	return oLhs.iData[0] == oRhs.iData[0];
}

/// check for equality between ShortStrings.
inline static bool operator==( const ShortString<8>& oLhs, const ShortString<8>& oRhs )
{
	return oLhs.iData[0] == oRhs.iData[0] && 
			 oLhs.iData[1] == oRhs.iData[1];
}

/// check for equality between ShortStrings.
inline static bool operator==( const ShortString<16>& oLhs, const ShortString<16>& oRhs )
{
	return oLhs.iData[0] == oRhs.iData[0] && 
			 oLhs.iData[1] == oRhs.iData[1] && 
			 oLhs.iData[2] == oRhs.iData[2] && 
			 oLhs.iData[3] == oRhs.iData[3];
}

/// check for equality between ShortStrings.
inline static bool operator==( const ShortString<32>& oLhs, const ShortString<32>& oRhs )
{
	return oLhs.iData[0] == oRhs.iData[0] && 
			 oLhs.iData[1] == oRhs.iData[1] && 
			 oLhs.iData[2] == oRhs.iData[2] && 
			 oLhs.iData[3] == oRhs.iData[3] && 
			 oLhs.iData[4] == oRhs.iData[4] && 
			 oLhs.iData[5] == oRhs.iData[5] && 
			 oLhs.iData[6] == oRhs.iData[6] && 
			 oLhs.iData[7] == oRhs.iData[7];
}

/// check for equality between ShortStrings.
inline static bool operator==( const ShortString<64>& oLhs, const ShortString<64>& oRhs )
{
	return oLhs.iData[0] == oRhs.iData[0] && 
			 oLhs.iData[1] == oRhs.iData[1] && 
			 oLhs.iData[2] == oRhs.iData[2] && 
			 oLhs.iData[3] == oRhs.iData[3] && 
			 oLhs.iData[4] == oRhs.iData[4] && 
			 oLhs.iData[5] == oRhs.iData[5] && 
			 oLhs.iData[6] == oRhs.iData[6] && 
			 oLhs.iData[7] == oRhs.iData[7] &&
			 oLhs.iData[8] == oRhs.iData[8] && 
			 oLhs.iData[9] == oRhs.iData[9] && 
			 oLhs.iData[10] == oRhs.iData[10] && 
			 oLhs.iData[11] == oRhs.iData[11] &&
			 oLhs.iData[12] == oRhs.iData[12] && 
			 oLhs.iData[13] == oRhs.iData[13] && 
			 oLhs.iData[14] == oRhs.iData[14] && 
			 oLhs.iData[15] == oRhs.iData[15];
}

/// check for equality between ShortStrings.
inline static bool operator==( const ShortString<128>& oLhs, const ShortString<128>& oRhs )
{
	return oLhs.iData[0] == oRhs.iData[0] && 
			 oLhs.iData[1] == oRhs.iData[1] && 
			 oLhs.iData[2] == oRhs.iData[2] && 
			 oLhs.iData[3] == oRhs.iData[3] && 
			 oLhs.iData[4] == oRhs.iData[4] && 
			 oLhs.iData[5] == oRhs.iData[5] && 
			 oLhs.iData[6] == oRhs.iData[6] && 
			 oLhs.iData[7] == oRhs.iData[7] &&
			 oLhs.iData[8] == oRhs.iData[8] && 
			 oLhs.iData[9] == oRhs.iData[9] && 
			 oLhs.iData[10] == oRhs.iData[10] && 
			 oLhs.iData[11] == oRhs.iData[11] &&
			 oLhs.iData[12] == oRhs.iData[12] && 
			 oLhs.iData[13] == oRhs.iData[13] && 
			 oLhs.iData[14] == oRhs.iData[14] && 
			 oLhs.iData[15] == oRhs.iData[15] &&
			 oLhs.iData[16] == oRhs.iData[16] && 
			 oLhs.iData[17] == oRhs.iData[17] &&
			 oLhs.iData[18] == oRhs.iData[18] && 
			 oLhs.iData[19] == oRhs.iData[19] && 
			 oLhs.iData[20] == oRhs.iData[20] && 
			 oLhs.iData[21] == oRhs.iData[21] &&
			 oLhs.iData[22] == oRhs.iData[22] && 
			 oLhs.iData[23] == oRhs.iData[23] &&
			 oLhs.iData[24] == oRhs.iData[24] && 
			 oLhs.iData[25] == oRhs.iData[25] && 
			 oLhs.iData[26] == oRhs.iData[26] && 
			 oLhs.iData[27] == oRhs.iData[27] &&
			 oLhs.iData[28] == oRhs.iData[28] && 
			 oLhs.iData[29] == oRhs.iData[29] && 
			 oLhs.iData[30] == oRhs.iData[30] && 
			 oLhs.iData[31] == oRhs.iData[31];
}

#endif