/*
   event horizon - lua/c++ event system demonstration
   Copyright (c) 2006 kevin meinert all rights reserved

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301  USA
   */

#ifndef PACKETSTUFF
#define PACKETSTUFF

#include <vector>
#include <stdlib.h>


/// base packet type, this is the readable data interface for packets
///   that you can store and pass around.
/// A description of the struct is inside the .attribs...
/// @see packet_ class for how to generate this data...
struct packet
{
   struct DataMember
   {
      enum DataType
      {
         TYPEINT, TYPESTR, TYPEFLOAT, TYPEBOOL
      };
      inline DataType map_to_type( bool val ) const { return TYPEBOOL; }
      inline DataType map_to_type( int val ) const { return TYPEINT; }
      inline DataType map_to_type( char val ) const { return TYPESTR; }
      inline DataType map_to_type( float val ) const { return TYPEFLOAT; }
      inline DataType map_to_type( bool val[] ) const { return TYPEBOOL; }
      inline DataType map_to_type( int val[] ) const { return TYPEINT; }
      inline DataType map_to_type( char val[] ) const { return TYPESTR; }
      inline DataType map_to_type( float val[] ) const { return TYPEFLOAT; }

      template <typename T>
      inline size_t sizesingleelt( T val ) const { return sizeof( T ); }
      template <typename T>
      inline size_t sizesingleelt( T val[] ) const { return sizeof( T ); }

      size_t location; //< offset to the datamember in the class/struct
      DataType type; //< what type is stored here?
      size_t length; //< number of elements in the datamember array
      size_t size; //< total size in bytes of data member
      size_t size_1; //< size of one element of datamember array
      const char* name; //< name of datamember

      template <typename PACKET_TYPE, typename D> 
      DataMember( const char* const n, D PACKET_TYPE::* val )
      { 
         PACKET_TYPE pt; 
         location = size_t( &( pt.*val ) ) - size_t( &pt ); 
         type = map_to_type( pt.*val ); 
         size = sizeof( pt.*val ); 
         size_1 = sizesingleelt( pt.*val ); 
         length = size / size_1; 
         name = n;
      }
   };

   const char* name;       //< name of class/struct type
   size_t size;            //< size of class/struct type
   std::vector<DataMember> attribs; //< attribute map of the type's data members
};

/// data packet description generator
/// usage:
/// packet o = 
///      packet_<mypacket>( "mypacket" )
///            .add( "int1", &mypacket::int1 )
///            .add( "int2", &mypacket::int2 )
///            .add( "str1", &mypacket::str1 )
///            .add( "str2", &mypacket::str2 );
/// read:
///      for (unsigned int x = 0; x < o.attribs.size(); ++x)
///      {
///         printf( "objsize:%d type:%d varsize:%d offset:%d\n", o.size, o.attribs[x].type, o.attribs[x].size, o.attribs[x].location );
///      }
template <typename PACKETTYPE>
struct packet_ : public packet
{
   packet_( const char* n )
   {
      name = n;
      size = sizeof( PACKETTYPE );
   }

   template <typename PACKET_TYPE, typename D> 
   packet_& add( const char* const name, D PACKET_TYPE::*val ) { attribs.push_back( packet::DataMember( name, val ) ); return *this; }
};

#endif