/*
   event horizon - lua/c++ event system demonstration
   Copyright (c) 2006 kevin meinert all rights reserved

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301  USA
   */

#include <iostream>
extern "C"
{
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
}
#include "EventSystem.h"

inline void safeStrCpy( char* dest, const char* src, int n )
{
   strncpy( dest, src, n );
   dest[n-1] = '\0';
}

/// this structure is the argument/infobuffer to the event
/// it is passed on the C++ side, and sent/recieved when an event is triggered 
struct TestStruct
{
   TestStruct() {}
   TestStruct( int b, float f, const char* o )
   {
      beans = b;
      frys = f;
      safeStrCpy( oompas, o, sizeof( oompas ) );
   }
   int beans;
   float frys;
   char oompas[20];
   static packet& getdef()
   {
      static packet o = 
         packet_<TestStruct>( "TestStruct" )
            .add( "beans", &TestStruct::beans )
            .add( "frys", &TestStruct::frys )
            .add( "oompas", &TestStruct::oompas );
      return o;
   }
};
/// event handler, we register this for the "fire" event
void trig( const void* arg )
{
   const TestStruct* t = (const TestStruct*)arg;
   std::cout << "------------------------" << std::endl;
   std::cout << "[c++]:" << std::endl;
   std::cout << "beans: " << t->beans << std::endl;
   std::cout << "frys: " << t->frys << std::endl;
   std::cout << "oompas: " << t->oompas << std::endl;
   std::cout << "------------------------" << std::endl;
}

/// for the [obj,member] event handler test
struct MyClass
{
   int my_id;
   void trig( const void* arg ) 
   { 
      const TestStruct* t = (const TestStruct*)arg;
      std::cout << "------------------------" << std::endl;
      std::cout << "[c++]:   we can reg [obj,memberfunc] pairs too!" << std::endl;
      std::cout << "this->my_id == " << my_id << std::endl;
      std::cout << "beans: " << t->beans << " frys: " << t->frys << " oompas: " << t->oompas << std::endl;
      std::cout << "------------------------" << std::endl;
   }
};

/// run some lua code...
int pcall_handler(lua_State* L) { std::cout << lua_tostring(L, -1) << std::endl; return 1; }
void dostring(lua_State* state, const char* const name, char const* str)
{
   lua_pushcclosure(state, &pcall_handler, 0);
   luaL_loadbuffer(state, str, std::strlen(str), name);
   lua_pcall(state, 0, 0, -2);
   lua_pop(state, 1);
}

/// entrypoint...
void main()
{
   lua_State* L = lua_open();      //< init lua
   luaL_openlibs( L );  //< add std libs to lua.
   EventSystem::instance().init( L );

   std::cout << "\n\nregister a lua event handler:" << std::endl;
   dostring( L, "script", 
      "function trig( beans, fryes, oompas )\n"
	   "  io.write( \"-------------------------\\n\" )\n"
      "  io.write( \"[lua]:\\n\" )\n"
	   "  io.write( \"beans: \" .. tostring( beans ) .. \"\\nfryes:\" .. tostring( fryes ) .. \"\\noompas:\" .. tostring( oompas ) .. \"\\n\" )\n"
	   "  io.write( \"-------------------------\\n\" )\n"
      "end\n"
	   "reg( \"fire\", trig )\n"
   );

   std::cout << "\n\nregister 2 lua (obj,member) event handlers:" << std::endl;
   dostring( L, "script", 
      "myclass = {}\n"
      "myclass.my_id = 39\n"
      "function myclass:trig( beans, fryes, oompas )\n"
	   "  io.write( \"-------------------------\\n\" )\n"
      "  io.write( \"[lua]:   we can reg [obj,memberfunc] pairs too!\\n\" )\n"
      "  io.write( \"self.my_id == \" .. tostring( self.my_id ) .. \"\\n\" )\n"
	   "  io.write( \"beans: \" .. tostring( beans ) .. \" fryes:\" .. tostring( fryes ) .. \" oompas:\" .. tostring( oompas ) .. \"\\n\" )\n"
	   "  io.write( \"-------------------------\\n\" )\n"
      "end\n"
      "reg( \"fire\", myclass.trig, myclass )\n"
      ""
      "myclass2 = {}\n"
      "myclass2.my_id = 987\n"
      "myclass2.trig = myclass.trig\n"
      "reg( \"fire\", myclass2.trig, myclass2 )\n"
   );

   std::cout << "\n\nregister a c++ event handler:" << std::endl;
   EventSystem::instance().reg( "fire", EventCallback( &trig, TestStruct::getdef() ) );

   std::cout << "\n\nregister 2 c++ [obj,member] event handler:" << std::endl;
   MyClass obj;
   obj.my_id = 98;
   EventSystem::instance().reg( "fire", EventCallback( &obj, &MyClass::trig, TestStruct::getdef() ) );
   MyClass obj2;
   obj2.my_id = 32;
   EventSystem::instance().reg( "fire", EventCallback( &obj2, &MyClass::trig, TestStruct::getdef() ) );

   // trigger:
   // here, we expect the string to be chopped off, there isn't room to store it in the packet...

   std::cout << "\n\ntrigger event from lua:" << std::endl;
   dostring( L, "script", 
      "trigger( \"fire\", 987654321, 0.112358, \"cake is for feasting, 33 times this morning\" )\n"
      //"trigger( \"fire\", 987654321.0, \"0.112358\", true )\n"
   );

   std::cout << "\n\ntrigger event from c++:" << std::endl;
   EventSystem::instance().trigger( "fire", TestStruct( 112358, 3.14159265f, "pi is good to taste, please use enough cornstarch" ) );
}